@@ -41,7 +41,7 @@ importance <- function(x, ...)  UseMethod("importance")
 ##' @aliases importance
 ##' @export 
 importance.ranger <- function(x, ...) {
-  if (class(x) != "ranger") {
+  if (!inherits(x, "ranger")) {
     stop("Object ist no ranger object.")
   }
   if (is.null(x$variable.importance) || length(x$variable.importance) < 1) {
@@ -76,8 +76,6 @@ importance.ranger <- function(x, ...) {
 ##' @param ... Further arguments passed to \code{ranger()}. Used in the "altmann" method only.
 ##' @return Variable importance and p-value for each variable.
 ##' @examples
-##' require(ranger)
-##' 
 ##' ## Janitza's p-values with corrected Gini importance
 ##' n <- 50
 ##' p <- 400
@@ -93,12 +91,12 @@ importance.ranger <- function(x, ...) {
 ##' @seealso \code{\link{ranger}}
 ##' @author Marvin N. Wright
 ##' @references
-##'   Janitza, S., Celik, E. & Boulesteix, A.-L., (2016). A computationally fast variable importance test for random forests for high-dimensional data. Adv Data Anal Classif \url{http://dx.doi.org/10.1007/s11634-016-0276-4}. \cr
+##'   Janitza, S., Celik, E. & Boulesteix, A.-L., (2016). A computationally fast variable importance test for random forests for high-dimensional data. Adv Data Anal Classif \doi{10.1007/s11634-016-0276-4}. \cr
 ##'   Altmann, A., Tolosi, L., Sander, O. & Lengauer, T. (2010). Permutation importance: a corrected feature importance measure, Bioinformatics 26:1340-1347.
 ##' @export 
 importance_pvalues <- function(x, method = c("janitza", "altmann"), num.permutations = 100, formula = NULL, data = NULL, ...) {
   method <- match.arg(method)
-  if (class(x) != "ranger" & class(x) != "holdoutRF") {
+  if (!inherits(x, c("ranger", "holdoutRF"))) {
     stop("Object is no ranger or holdoutRF object.")
   }
   if (x$importance.mode == "none" || is.null(x$variable.importance) || length(x$variable.importance) < 1) {
@@ -109,7 +107,7 @@ importance_pvalues <- function(x, method = c("janitza", "altmann"), num.permutat
     if (x$importance.mode == "impurity") {
       stop("Impurity variable importance found. Please use (hold-out) permutation importance or corrected impurity importance to use this method.")
     }
-    if (class(x) != "holdoutRF" && x$importance.mode == "permutation") {
+    if (!inherits(x, "holdoutRF") && x$importance.mode == "permutation") {
       warning("Permutation variable importance found, inaccurate p-values. Please use hold-out permutation importance or corrected impurity importance to use this method.")
     }
     if (x$treetype != "Classification") {
@@ -133,7 +131,7 @@ importance_pvalues <- function(x, method = c("janitza", "altmann"), num.permutat
       warning("Only few negative importance values found, inaccurate p-values. Consider the 'altmann' approach.")
     }
   } else if (method == "altmann") {
-    if (class(x) != "ranger") {
+    if (!inherits(x, "ranger")) {
       stop("Altmann method not available for holdoutRF objects.")
     }
     if (is.null(formula) || is.null(data)) {
