@@ -29,40 +29,15 @@
 namespace ranger {
 
 Forest::Forest() :
-    verbose_out(0), num_trees(DEFAULT_NUM_TREE), mtry(0), min_node_size(0), num_variables(0), num_independent_variables(
-        0), seed(0), dependent_varID(0), num_samples(0), prediction_mode(false), memory_mode(MEM_DOUBLE), sample_with_replacement(
-        true), memory_saving_splitting(false), splitrule(DEFAULT_SPLITRULE), predict_all(false), keep_inbag(false), sample_fraction(
-        { 1 }), holdout(false), prediction_type(DEFAULT_PREDICTIONTYPE), num_random_splits(DEFAULT_NUM_RANDOM_SPLITS), max_depth(
+    verbose_out(0), num_trees(DEFAULT_NUM_TREE), mtry(0), min_node_size(0), num_independent_variables(0), seed(0), num_samples(
+        0), prediction_mode(false), memory_mode(MEM_DOUBLE), sample_with_replacement(true), memory_saving_splitting(
+        false), splitrule(DEFAULT_SPLITRULE), predict_all(false), keep_inbag(false), sample_fraction( { 1 }), holdout(
+        false), prediction_type(DEFAULT_PREDICTIONTYPE), num_random_splits(DEFAULT_NUM_RANDOM_SPLITS), max_depth(
         DEFAULT_MAXDEPTH), alpha(DEFAULT_ALPHA), minprop(DEFAULT_MINPROP), num_threads(DEFAULT_NUM_THREADS), data { }, overall_prediction_error(
-    NAN), importance_mode(DEFAULT_IMPORTANCE_MODE), progress(0) {
+    NAN), importance_mode(DEFAULT_IMPORTANCE_MODE), regularization_usedepth(false), progress(0) {
 }
 
 // #nocov start
-std::unique_ptr<Data> load_data_from_file(const std::string& data_path, const MemoryMode memory_mode,
-    std::ostream* verbose_out = nullptr) {
-  std::unique_ptr<Data> result { };
-  switch (memory_mode) {
-  case MEM_DOUBLE:
-    result = make_unique<DataDouble>();
-    break;
-  case MEM_FLOAT:
-    result = make_unique<DataFloat>();
-    break;
-  case MEM_CHAR:
-    result = make_unique<DataChar>();
-    break;
-  }
-
-  if (verbose_out)
-    *verbose_out << "Loading input file: " << data_path << "." << std::endl;
-  bool found_rounding_error = result->loadFromFile(data_path);
-  if (found_rounding_error && verbose_out) {
-    *verbose_out << "Warning: Rounding or Integer overflow occurred. Use FLOAT or DOUBLE precision to avoid this."
-        << std::endl;
-  }
-  return result;
-}
-
 void Forest::initCpp(std::string dependent_variable_name, MemoryMode memory_mode, std::string input_file, uint mtry,
     std::string output_prefix, uint num_trees, std::ostream* verbose_out, uint seed, uint num_threads,
     std::string load_forest_filename, ImportanceMode importance_mode, uint min_node_size,
@@ -70,25 +45,45 @@ void Forest::initCpp(std::string dependent_variable_name, MemoryMode memory_mode
     std::string status_variable_name, bool sample_with_replacement,
     const std::vector<std::string>& unordered_variable_names, bool memory_saving_splitting, SplitRule splitrule,
     std::string case_weights_file, bool predict_all, double sample_fraction, double alpha, double minprop, bool holdout,
-    PredictionType prediction_type, uint num_random_splits, uint max_depth) {
+    PredictionType prediction_type, uint num_random_splits, uint max_depth,
+    const std::vector<double>& regularization_factor, bool regularization_usedepth) {
 
+  this->memory_mode = memory_mode;
   this->verbose_out = verbose_out;
 
+  if (!dependent_variable_name.empty()) {
+    if (status_variable_name.empty()) {
+      this->dependent_variable_names = {dependent_variable_name};
+    } else {
+      this->dependent_variable_names = {dependent_variable_name, status_variable_name};
+    }
+  }
+
   // Set prediction mode
   bool prediction_mode = false;
   if (!load_forest_filename.empty()) {
     prediction_mode = true;
   }
 
-  // Sample fraction to vector
+  // Sample fraction default and convert to vector
+  if (sample_fraction == 0) {
+    if (sample_with_replacement) {
+      sample_fraction = DEFAULT_SAMPLE_FRACTION_REPLACE;
+    } else {
+      sample_fraction = DEFAULT_SAMPLE_FRACTION_NOREPLACE;
+    }
+  }
   std::vector<double> sample_fraction_vector = { sample_fraction };
 
+  if (prediction_mode) {
+    loadDependentVariableNamesFromFile(load_forest_filename);
+  }
+
   // Call other init function
-  init(dependent_variable_name, memory_mode, load_data_from_file(input_file, memory_mode, verbose_out), mtry,
-      output_prefix, num_trees, seed, num_threads, importance_mode, min_node_size, status_variable_name,
-      prediction_mode, sample_with_replacement, unordered_variable_names, memory_saving_splitting, splitrule,
-      predict_all, sample_fraction_vector, alpha, minprop, holdout, prediction_type, num_random_splits, false,
-      max_depth);
+  init(loadDataFromFile(input_file), mtry, output_prefix, num_trees, seed, num_threads, importance_mode,
+      min_node_size, prediction_mode, sample_with_replacement, unordered_variable_names, memory_saving_splitting,
+      splitrule, predict_all, sample_fraction_vector, alpha, minprop, holdout, prediction_type, num_random_splits,
+      false, max_depth, regularization_factor, regularization_usedepth);
 
   if (prediction_mode) {
     loadFromFile(load_forest_filename);
@@ -104,7 +99,7 @@ void Forest::initCpp(std::string dependent_variable_name, MemoryMode memory_mode
     std::vector<std::vector<double>> split_select_weights;
     split_select_weights.resize(1);
     loadDoubleVectorFromFile(split_select_weights[0], split_select_weights_file);
-    if (split_select_weights[0].size() != num_variables - 1) {
+    if (split_select_weights[0].size() != num_independent_variables) {
       throw std::runtime_error("Number of split select weights is not equal to number of independent variables.");
     }
     setSplitWeightVector(split_select_weights);
@@ -139,22 +134,23 @@ void Forest::initCpp(std::string dependent_variable_name, MemoryMode memory_mode
 }
 // #nocov end
 
-void Forest::initR(std::string dependent_variable_name, std::unique_ptr<Data> input_data, uint mtry, uint num_trees,
-    std::ostream* verbose_out, uint seed, uint num_threads, ImportanceMode importance_mode, uint min_node_size,
+void Forest::initR(std::unique_ptr<Data> input_data, uint mtry, uint num_trees, std::ostream* verbose_out, uint seed,
+    uint num_threads, ImportanceMode importance_mode, uint min_node_size,
     std::vector<std::vector<double>>& split_select_weights, const std::vector<std::string>& always_split_variable_names,
-    std::string status_variable_name, bool prediction_mode, bool sample_with_replacement,
-    const std::vector<std::string>& unordered_variable_names, bool memory_saving_splitting, SplitRule splitrule,
-    std::vector<double>& case_weights, std::vector<std::vector<size_t>>& manual_inbag, bool predict_all,
-    bool keep_inbag, std::vector<double>& sample_fraction, double alpha, double minprop, bool holdout,
-    PredictionType prediction_type, uint num_random_splits, bool order_snps, uint max_depth) {
+    bool prediction_mode, bool sample_with_replacement, const std::vector<std::string>& unordered_variable_names,
+    bool memory_saving_splitting, SplitRule splitrule, std::vector<double>& case_weights,
+    std::vector<std::vector<size_t>>& manual_inbag, bool predict_all, bool keep_inbag,
+    std::vector<double>& sample_fraction, double alpha, double minprop, bool holdout, PredictionType prediction_type,
+    uint num_random_splits, bool order_snps, uint max_depth, const std::vector<double>& regularization_factor,
+    bool regularization_usedepth) {
 
   this->verbose_out = verbose_out;
 
   // Call other init function
-  init(dependent_variable_name, MEM_DOUBLE, std::move(input_data), mtry, "", num_trees, seed, num_threads,
-      importance_mode, min_node_size, status_variable_name, prediction_mode, sample_with_replacement,
-      unordered_variable_names, memory_saving_splitting, splitrule, predict_all, sample_fraction, alpha, minprop,
-      holdout, prediction_type, num_random_splits, order_snps, max_depth);
+  init(std::move(input_data), mtry, "", num_trees, seed, num_threads, importance_mode, min_node_size,
+      prediction_mode, sample_with_replacement, unordered_variable_names, memory_saving_splitting, splitrule,
+      predict_all, sample_fraction, alpha, minprop, holdout, prediction_type, num_random_splits, order_snps, max_depth,
+      regularization_factor, regularization_usedepth);
 
   // Set variables to be always considered for splitting
   if (!always_split_variable_names.empty()) {
@@ -183,12 +179,12 @@ void Forest::initR(std::string dependent_variable_name, std::unique_ptr<Data> in
   this->keep_inbag = keep_inbag;
 }
 
-void Forest::init(std::string dependent_variable_name, MemoryMode memory_mode, std::unique_ptr<Data> input_data,
-    uint mtry, std::string output_prefix, uint num_trees, uint seed, uint num_threads, ImportanceMode importance_mode,
-    uint min_node_size, std::string status_variable_name, bool prediction_mode, bool sample_with_replacement,
-    const std::vector<std::string>& unordered_variable_names, bool memory_saving_splitting, SplitRule splitrule,
-    bool predict_all, std::vector<double>& sample_fraction, double alpha, double minprop, bool holdout,
-    PredictionType prediction_type, uint num_random_splits, bool order_snps, uint max_depth) {
+void Forest::init(std::unique_ptr<Data> input_data, uint mtry, std::string output_prefix,
+    uint num_trees, uint seed, uint num_threads, ImportanceMode importance_mode, uint min_node_size,
+    bool prediction_mode, bool sample_with_replacement, const std::vector<std::string>& unordered_variable_names,
+    bool memory_saving_splitting, SplitRule splitrule, bool predict_all, std::vector<double>& sample_fraction,
+    double alpha, double minprop, bool holdout, PredictionType prediction_type, uint num_random_splits, bool order_snps,
+    uint max_depth, const std::vector<double>& regularization_factor, bool regularization_usedepth) {
 
   // Initialize data with memmode
   this->data = std::move(input_data);
@@ -219,7 +215,6 @@ void Forest::init(std::string dependent_variable_name, MemoryMode memory_mode, s
   this->output_prefix = output_prefix;
   this->importance_mode = importance_mode;
   this->min_node_size = min_node_size;
-  this->memory_mode = memory_mode;
   this->prediction_mode = prediction_mode;
   this->sample_with_replacement = sample_with_replacement;
   this->memory_saving_splitting = memory_saving_splitting;
@@ -232,26 +227,19 @@ void Forest::init(std::string dependent_variable_name, MemoryMode memory_mode, s
   this->prediction_type = prediction_type;
   this->num_random_splits = num_random_splits;
   this->max_depth = max_depth;
+  this->regularization_factor = regularization_factor;
+  this->regularization_usedepth = regularization_usedepth;
 
   // Set number of samples and variables
   num_samples = data->getNumRows();
-  num_variables = data->getNumCols();
-
-  // Convert dependent variable name to ID
-  if (!prediction_mode && !dependent_variable_name.empty()) {
-    dependent_varID = data->getVariableID(dependent_variable_name);
-  }
+  num_independent_variables = data->getNumCols();
 
   // Set unordered factor variables
   if (!prediction_mode) {
     data->setIsOrderedVariable(unordered_variable_names);
   }
 
-  data->addNoSplitVariable(dependent_varID);
-
-  initInternal(status_variable_name);
-
-  num_independent_variables = num_variables - data->getNoSplitVariables().size();
+  initInternal();
 
   // Init split select weights
   split_select_weights.push_back(std::vector<double>());
@@ -260,7 +248,7 @@ void Forest::init(std::string dependent_variable_name, MemoryMode memory_mode, s
   manual_inbag.push_back(std::vector<size_t>());
 
   // Check if mtry is in valid range
-  if (this->mtry > num_variables - 1) {
+  if (this->mtry > num_independent_variables) {
     throw std::runtime_error("mtry can not be larger than number of variables in data.");
   }
 
@@ -276,7 +264,20 @@ void Forest::init(std::string dependent_variable_name, MemoryMode memory_mode, s
 
   // Order SNP levels if in "order" splitting
   if (!prediction_mode && order_snps) {
-    data->orderSnpLevels(dependent_variable_name, (importance_mode == IMP_GINI_CORRECTED));
+    data->orderSnpLevels((importance_mode == IMP_GINI_CORRECTED));
+  }
+
+  // Regularization
+  if (regularization_factor.size() > 0) {
+    if (regularization_factor.size() == 1 && num_independent_variables > 1) {
+      double single_regularization_factor = regularization_factor[0];
+      this->regularization_factor.resize(num_independent_variables, single_regularization_factor);
+    } else if (regularization_factor.size() != num_independent_variables) {
+      throw std::runtime_error("Use 1 or p (the number of predictor variables) regularization factors.");
+    }
+
+    // Set all variables to not used
+    split_varIDs_used.resize(num_independent_variables, false);
   }
 }
 
@@ -302,7 +303,8 @@ void Forest::run(bool verbose, bool compute_oob_error) {
       computePredictionError();
     }
 
-    if (importance_mode == IMP_PERM_BREIMAN || importance_mode == IMP_PERM_LIAW || importance_mode == IMP_PERM_RAW) {
+    if (importance_mode == IMP_PERM_BREIMAN || importance_mode == IMP_PERM_LIAW || importance_mode == IMP_PERM_RAW
+        || importance_mode == IMP_PERM_CASEWISE) {
       if (verbose && verbose_out) {
         *verbose_out << "Computing permutation variable importance .." << std::endl;
       }
@@ -318,8 +320,9 @@ void Forest::writeOutput() {
     *verbose_out << std::endl;
   writeOutputInternal();
   if (verbose_out) {
-    *verbose_out << "Dependent variable name:           " << data->getVariableNames()[dependent_varID] << std::endl;
-    *verbose_out << "Dependent variable ID:             " << dependent_varID << std::endl;
+    if (dependent_variable_names.size() >= 1) {
+      *verbose_out << "Dependent variable name:           " << dependent_variable_names[0] << std::endl;
+    }
     *verbose_out << "Number of trees:                   " << num_trees << std::endl;
     *verbose_out << "Sample size:                       " << num_samples << std::endl;
     *verbose_out << "Number of independent variables:   " << num_independent_variables << std::endl;
@@ -340,7 +343,7 @@ void Forest::writeOutput() {
       *verbose_out << std::endl;
     }
 
-    if (!split_select_weights.empty() & !split_select_weights[0].empty()) {
+    if (!split_select_weights.empty() && !split_select_weights[0].empty()) {
       if (verbose_out) {
         *verbose_out
             << "Warning: Split select weights used. Variable importance measures are only comparable for variables with equal weights."
@@ -366,16 +369,29 @@ void Forest::writeImportanceFile() {
     throw std::runtime_error("Could not write to importance file: " + filename + ".");
   }
 
-  // Write importance to file
-  for (size_t i = 0; i < variable_importance.size(); ++i) {
-    size_t varID = i;
-    for (auto& skip : data->getNoSplitVariables()) {
-      if (varID >= skip) {
-        ++varID;
+  if (importance_mode == IMP_PERM_CASEWISE) {
+    // Write variable names
+    for (auto& variable_name : data->getVariableNames()) {
+      importance_file << variable_name << " ";
+    }
+    importance_file << std::endl;
+
+    // Write importance values
+    for (size_t i = 0; i < num_samples; ++i) {
+      for (size_t j = 0; j < num_independent_variables; ++j) {
+        if (variable_importance_casewise.size() <= (j * num_samples + i)) {
+          throw std::runtime_error("Memory error in local variable importance.");
+        }
+        importance_file << variable_importance_casewise[j * num_samples + i] << " ";
       }
+      importance_file << std::endl;
+    }
+  } else {
+    // Write importance to file
+    for (size_t i = 0; i < variable_importance.size(); ++i) {
+      std::string variable_name = data->getVariableNames()[i];
+      importance_file << variable_name << ": " << variable_importance[i] << std::endl;
     }
-    std::string variable_name = data->getVariableNames()[varID];
-    importance_file << variable_name << ": " << variable_importance[i] << std::endl;
   }
 
   importance_file.close();
@@ -393,8 +409,18 @@ void Forest::saveToFile() {
     throw std::runtime_error("Could not write to output file: " + filename + ".");
   }
 
-  // Write dependent_varID
-  outfile.write((char*) &dependent_varID, sizeof(dependent_varID));
+  // Write dependent variable names
+  uint num_dependent_variables = dependent_variable_names.size();
+  if (num_dependent_variables >= 1) {
+    outfile.write((char*) &num_dependent_variables, sizeof(num_dependent_variables));
+    for (auto& var_name : dependent_variable_names) {
+      size_t length = var_name.size();
+      outfile.write((char*) &length, sizeof(length));
+      outfile.write((char*) var_name.c_str(), length * sizeof(char));
+    }
+  } else {
+    throw std::runtime_error("Missing dependent variable name.");
+  }
 
   // Write num_trees
   outfile.write((char*) &num_trees, sizeof(num_trees));
@@ -450,17 +476,18 @@ void Forest::grow() {
       tree_manual_inbag = &manual_inbag[0];
     }
 
-    trees[i]->init(data.get(), mtry, dependent_varID, num_samples, tree_seed, &deterministic_varIDs,
-        &split_select_varIDs, tree_split_select_weights, importance_mode, min_node_size, sample_with_replacement,
-        memory_saving_splitting, splitrule, &case_weights, tree_manual_inbag, keep_inbag, &sample_fraction, alpha,
-        minprop, holdout, num_random_splits, max_depth);
+    trees[i]->init(data.get(), mtry, num_samples, tree_seed, &deterministic_varIDs, tree_split_select_weights,
+        importance_mode, min_node_size, sample_with_replacement, memory_saving_splitting, splitrule, &case_weights,
+        tree_manual_inbag, keep_inbag, &sample_fraction, alpha, minprop, holdout, num_random_splits, max_depth,
+        &regularization_factor, regularization_usedepth, &split_varIDs_used);
   }
 
-// Init variable importance
+  // Init variable importance
   variable_importance.resize(num_independent_variables, 0);
 
-// Grow trees in multiple threads
+  // Grow trees in multiple threads
 #ifdef OLD_WIN_R_BUILD
+  // #nocov start
   progress = 0;
   clock_t start_time = clock();
   clock_t lap_time = clock();
@@ -469,6 +496,7 @@ void Forest::grow() {
     progress++;
     showProgress("Growing trees..", start_time, lap_time);
   }
+  // #nocov end
 #else
   progress = 0;
 #ifdef R_BUILD
@@ -479,7 +507,7 @@ void Forest::grow() {
   std::vector<std::thread> threads;
   threads.reserve(num_threads);
 
-// Initailize importance per thread
+  // Initialize importance per thread
   std::vector<std::vector<double>> variable_importance_threads(num_threads);
 
   for (uint i = 0; i < num_threads; ++i) {
@@ -512,7 +540,7 @@ void Forest::grow() {
 
 #endif
 
-// Divide importance by number of trees
+  // Divide importance by number of trees
   if (importance_mode == IMP_GINI || importance_mode == IMP_GINI_CORRECTED) {
     for (auto& v : variable_importance) {
       v /= num_trees;
@@ -522,8 +550,9 @@ void Forest::grow() {
 
 void Forest::predict() {
 
-// Predict trees in multiple threads and join the threads with the main thread
+  // Predict trees in multiple threads and join the threads with the main thread
 #ifdef OLD_WIN_R_BUILD
+  // #nocov start
   progress = 0;
   clock_t start_time = clock();
   clock_t lap_time = clock();
@@ -538,6 +567,7 @@ void Forest::predict() {
   for (size_t sample_idx = 0; sample_idx < data->getNumRows(); ++sample_idx) {
     predictInternal(sample_idx);
   }
+  // #nocov end
 #else
   progress = 0;
 #ifdef R_BUILD
@@ -579,8 +609,9 @@ void Forest::predict() {
 
 void Forest::computePredictionError() {
 
-// Predict trees in multiple threads
+  // Predict trees in multiple threads
 #ifdef OLD_WIN_R_BUILD
+  // #nocov start
   progress = 0;
   clock_t start_time = clock();
   clock_t lap_time = clock();
@@ -589,6 +620,7 @@ void Forest::computePredictionError() {
     progress++;
     showProgress("Predicting..", start_time, lap_time);
   }
+  // #nocov end
 #else
   std::vector<std::thread> threads;
   threads.reserve(num_threads);
@@ -614,25 +646,30 @@ void Forest::computePredictionError() {
 
 void Forest::computePermutationImportance() {
 
-// Compute tree permutation importance in multiple threads
+  // Compute tree permutation importance in multiple threads
 #ifdef OLD_WIN_R_BUILD
+  // #nocov start
   progress = 0;
   clock_t start_time = clock();
   clock_t lap_time = clock();
 
-// Initailize importance and variance
+  // Initialize importance and variance
   variable_importance.resize(num_independent_variables, 0);
   std::vector<double> variance;
   if (importance_mode == IMP_PERM_BREIMAN || importance_mode == IMP_PERM_LIAW) {
     variance.resize(num_independent_variables, 0);
   }
+  if (importance_mode == IMP_PERM_CASEWISE) {
+    variable_importance_casewise.resize(num_independent_variables * num_samples, 0);
+  }
 
-// Compute importance
+  // Compute importance
   for (size_t i = 0; i < num_trees; ++i) {
-    trees[i]->computePermutationImportance(variable_importance, variance);
+    trees[i]->computePermutationImportance(variable_importance, variance, variable_importance_casewise);
     progress++;
     showProgress("Computing permutation importance..", start_time, lap_time);
   }
+
 #else
   progress = 0;
 #ifdef R_BUILD
@@ -643,18 +680,23 @@ void Forest::computePermutationImportance() {
   std::vector<std::thread> threads;
   threads.reserve(num_threads);
 
-// Initailize importance and variance
+  // Initialize importance and variance
   std::vector<std::vector<double>> variable_importance_threads(num_threads);
   std::vector<std::vector<double>> variance_threads(num_threads);
+  std::vector<std::vector<double>> variable_importance_casewise_threads(num_threads);
 
-// Compute importance
+  // Compute importance
   for (uint i = 0; i < num_threads; ++i) {
     variable_importance_threads[i].resize(num_independent_variables, 0);
     if (importance_mode == IMP_PERM_BREIMAN || importance_mode == IMP_PERM_LIAW) {
       variance_threads[i].resize(num_independent_variables, 0);
     }
+    if (importance_mode == IMP_PERM_CASEWISE) {
+      variable_importance_casewise_threads[i].resize(num_independent_variables * num_samples, 0);
+    }
     threads.emplace_back(&Forest::computeTreePermutationImportanceInThread, this, i,
-        std::ref(variable_importance_threads[i]), std::ref(variance_threads[i]));
+        std::ref(variable_importance_threads[i]), std::ref(variance_threads[i]),
+        std::ref(variable_importance_casewise_threads[i]));
   }
   showProgress("Computing permutation importance..", num_trees);
   for (auto &thread : threads) {
@@ -667,7 +709,7 @@ void Forest::computePermutationImportance() {
   }
 #endif
 
-// Sum thread importances
+  // Sum thread importances
   variable_importance.resize(num_independent_variables, 0);
   for (size_t i = 0; i < num_independent_variables; ++i) {
     for (uint j = 0; j < num_threads; ++j) {
@@ -676,7 +718,7 @@ void Forest::computePermutationImportance() {
   }
   variable_importance_threads.clear();
 
-// Sum thread variances
+  // Sum thread variances
   std::vector<double> variance(num_independent_variables, 0);
   if (importance_mode == IMP_PERM_BREIMAN || importance_mode == IMP_PERM_LIAW) {
     for (size_t i = 0; i < num_independent_variables; ++i) {
@@ -686,6 +728,17 @@ void Forest::computePermutationImportance() {
     }
     variance_threads.clear();
   }
+
+  // Sum thread casewise importances
+  if (importance_mode == IMP_PERM_CASEWISE) {
+    variable_importance_casewise.resize(num_independent_variables * num_samples, 0);
+    for (size_t i = 0; i < variable_importance_casewise.size(); ++i) {
+      for (uint j = 0; j < num_threads; ++j) {
+        variable_importance_casewise[i] += variable_importance_casewise_threads[j][i];
+      }
+    }
+    variable_importance_casewise_threads.clear();
+  }
 #endif
 
   for (size_t i = 0; i < variable_importance.size(); ++i) {
@@ -699,6 +752,12 @@ void Forest::computePermutationImportance() {
       }
     }
   }
+
+  if (importance_mode == IMP_PERM_CASEWISE) {
+    for (size_t i = 0; i < variable_importance_casewise.size(); ++i) {
+      variable_importance_casewise[i] /= num_trees;
+    }
+  }
 }
 
 #ifndef OLD_WIN_R_BUILD
@@ -776,10 +835,10 @@ void Forest::predictInternalInThread(uint thread_idx) {
 }
 
 void Forest::computeTreePermutationImportanceInThread(uint thread_idx, std::vector<double>& importance,
-    std::vector<double>& variance) {
+    std::vector<double>& variance, std::vector<double>& importance_casewise) {
   if (thread_ranges.size() > thread_idx + 1) {
     for (size_t i = thread_ranges[thread_idx]; i < thread_ranges[thread_idx + 1]; ++i) {
-      trees[i]->computePermutationImportance(importance, variance);
+      trees[i]->computePermutationImportance(importance, variance, importance_casewise);
 
       // Check for user interrupt
 #ifdef R_BUILD
@@ -805,38 +864,96 @@ void Forest::loadFromFile(std::string filename) {
   if (verbose_out)
     *verbose_out << "Loading forest from file " << filename << "." << std::endl;
 
-// Open file for reading
+  // Open file for reading
   std::ifstream infile;
   infile.open(filename, std::ios::binary);
   if (!infile.good()) {
     throw std::runtime_error("Could not read from input file: " + filename + ".");
   }
 
-// Read dependent_varID and num_trees
-  infile.read((char*) &dependent_varID, sizeof(dependent_varID));
+  // Skip dependent variable names (already read)
+  uint num_dependent_variables;
+  infile.read((char*) &num_dependent_variables, sizeof(num_dependent_variables));
+  for (size_t i = 0; i < num_dependent_variables; ++i) {
+    size_t length;
+    infile.read((char*) &length, sizeof(size_t));
+    infile.ignore(length);
+  }
+
+  // Read num_trees
   infile.read((char*) &num_trees, sizeof(num_trees));
 
-// Read is_ordered_variable
+  // Read is_ordered_variable
   readVector1D(data->getIsOrderedVariable(), infile);
 
-// Read tree data. This is different for tree types -> virtual function
+  // Read tree data. This is different for tree types -> virtual function
   loadFromFileInternal(infile);
 
   infile.close();
 
-// Create thread ranges
+  // Create thread ranges
   equalSplit(thread_ranges, 0, num_trees - 1, num_threads);
 }
+
+void Forest::loadDependentVariableNamesFromFile(std::string filename) {
+
+  // Open file for reading
+  std::ifstream infile;
+  infile.open(filename, std::ios::binary);
+  if (!infile.good()) {
+    throw std::runtime_error("Could not read from input file: " + filename + ".");
+  }
+
+  // Read dependent variable names
+  dependent_variable_names.clear();
+  uint num_dependent_variables = 0;
+  infile.read((char*) &num_dependent_variables, sizeof(num_dependent_variables));
+  for (size_t i = 0; i < num_dependent_variables; ++i) {
+    size_t length;
+    infile.read((char*) &length, sizeof(size_t));
+    char* temp = new char[length + 1];
+    infile.read((char*) temp, length * sizeof(char));
+    temp[length] = '\0';
+    dependent_variable_names.push_back(temp);
+    delete[] temp;
+  }
+
+  infile.close();
+}
+
+std::unique_ptr<Data> Forest::loadDataFromFile(const std::string& data_path) {
+  std::unique_ptr<Data> result { };
+  switch (memory_mode) {
+  case MEM_DOUBLE:
+    result = make_unique<DataDouble>();
+    break;
+  case MEM_FLOAT:
+    result = make_unique<DataFloat>();
+    break;
+  case MEM_CHAR:
+    result = make_unique<DataChar>();
+    break;
+  }
+
+  if (verbose_out)
+    *verbose_out << "Loading input file: " << data_path << "." << std::endl;
+  bool found_rounding_error = result->loadFromFile(data_path, dependent_variable_names);
+  if (found_rounding_error && verbose_out) {
+    *verbose_out << "Warning: Rounding or Integer overflow occurred. Use FLOAT or DOUBLE precision to avoid this."
+        << std::endl;
+  }
+  return result;
+}
 // #nocov end
 
 void Forest::setSplitWeightVector(std::vector<std::vector<double>>& split_select_weights) {
 
-// Size should be 1 x num_independent_variables or num_trees x num_independent_variables
+  // Size should be 1 x num_independent_variables or num_trees x num_independent_variables
   if (split_select_weights.size() != 1 && split_select_weights.size() != num_trees) {
     throw std::runtime_error("Size of split select weights not equal to 1 or number of trees.");
   }
 
-// Reserve space
+  // Reserve space
   size_t num_weights = num_independent_variables;
   if (importance_mode == IMP_GINI_CORRECTED) {
     num_weights = 2 * num_independent_variables;
@@ -847,12 +964,16 @@ void Forest::setSplitWeightVector(std::vector<std::vector<double>>& split_select
     this->split_select_weights.clear();
     this->split_select_weights.resize(num_trees, std::vector<double>(num_weights));
   }
-  this->split_select_varIDs.resize(num_weights);
-  deterministic_varIDs.reserve(num_weights);
+  
+  // Deterministic varIDs 
+  std::vector<bool> is_deterministic(num_weights, false);
+  for (size_t i = 0; i < deterministic_varIDs.size(); ++i) {
+    is_deterministic[i] = true;
+  }
 
   // Split up in deterministic and weighted variables, ignore zero weights
-  size_t num_zero_weights = 0;
   for (size_t i = 0; i < split_select_weights.size(); ++i) {
+    size_t num_zero_weights = 0;
 
     // Size should be 1 x num_independent_variables or num_trees x num_independent_variables
     if (split_select_weights[i].size() != num_independent_variables) {
@@ -862,31 +983,12 @@ void Forest::setSplitWeightVector(std::vector<std::vector<double>>& split_select
     for (size_t j = 0; j < split_select_weights[i].size(); ++j) {
       double weight = split_select_weights[i][j];
 
-      if (i == 0) {
-        size_t varID = j;
-        for (auto& skip : data->getNoSplitVariables()) {
-          if (varID >= skip) {
-            ++varID;
-          }
-        }
-
-        if (weight == 1) {
-          deterministic_varIDs.push_back(varID);
-        } else if (weight < 1 && weight > 0) {
-          this->split_select_varIDs[j] = varID;
-          this->split_select_weights[i][j] = weight;
-        } else if (weight == 0) {
-          ++num_zero_weights;
-        } else if (weight < 0 || weight > 1) {
-          throw std::runtime_error("One or more split select weights not in range [0,1].");
-        }
-
+      if (weight == 0 || is_deterministic[j]) {
+        ++num_zero_weights;
+      } else if (weight < 0 || weight > 1) {
+        throw std::runtime_error("One or more split select weights not in range [0,1].");
       } else {
-        if (weight < 1 && weight > 0) {
-          this->split_select_weights[i][j] = weight;
-        } else if (weight < 0 || weight > 1) {
-          throw std::runtime_error("One or more split select weights not in range [0,1].");
-        }
+        this->split_select_weights[i][j] = weight;
       }
     }
 
@@ -894,26 +996,11 @@ void Forest::setSplitWeightVector(std::vector<std::vector<double>>& split_select
     if (importance_mode == IMP_GINI_CORRECTED) {
       std::vector<double>* sw = &(this->split_select_weights[i]);
       std::copy_n(sw->begin(), num_independent_variables, sw->begin() + num_independent_variables);
-
-      for (size_t k = 0; k < num_independent_variables; ++k) {
-        split_select_varIDs[num_independent_variables + k] = num_variables + k;
-      }
-
-      size_t num_deterministic_varIDs = deterministic_varIDs.size();
-      for (size_t k = 0; k < num_deterministic_varIDs; ++k) {
-        size_t varID = deterministic_varIDs[k];
-        for (auto& skip : data->getNoSplitVariables()) {
-          if (varID >= skip) {
-            --varID;
-          }
-        }
-        deterministic_varIDs.push_back(varID + num_variables);
-      }
     }
-  }
 
-  if (num_weights - deterministic_varIDs.size() - num_zero_weights < mtry) {
-    throw std::runtime_error("Too many zeros or ones in split select weights. Need at least mtry variables to split at.");
+    if (num_weights - num_zero_weights < mtry) {
+      throw std::runtime_error("Too many zeros in split select weights. Need at least mtry variables to split at.");
+    }
   }
 }
 
@@ -935,21 +1022,16 @@ void Forest::setAlwaysSplitVariables(const std::vector<std::string>& always_spli
   if (importance_mode == IMP_GINI_CORRECTED) {
     size_t num_deterministic_varIDs = deterministic_varIDs.size();
     for (size_t k = 0; k < num_deterministic_varIDs; ++k) {
-      size_t varID = deterministic_varIDs[k];
-      for (auto& skip : data->getNoSplitVariables()) {
-        if (varID >= skip) {
-          --varID;
-        }
-      }
-      deterministic_varIDs.push_back(varID + num_variables);
+      deterministic_varIDs.push_back(k + num_independent_variables);
     }
   }
 }
 
 #ifdef OLD_WIN_R_BUILD
+// #nocov start
 void Forest::showProgress(std::string operation, clock_t start_time, clock_t& lap_time) {
 
-// Check for user interrupt
+  // Check for user interrupt
   if (checkInterrupt()) {
     throw std::runtime_error("User interrupt.");
   }
@@ -966,6 +1048,7 @@ void Forest::showProgress(std::string operation, clock_t start_time, clock_t& la
     lap_time = clock();
   }
 }
+// #nocov end
 #else
 void Forest::showProgress(std::string operation, size_t max_progress) {
   using std::chrono::steady_clock;
@@ -976,7 +1059,7 @@ void Forest::showProgress(std::string operation, size_t max_progress) {
   steady_clock::time_point last_time = steady_clock::now();
   std::unique_lock<std::mutex> lock(mutex);
 
-// Wait for message from threads and show output if enough time elapsed
+  // Wait for message from threads and show output if enough time elapsed
   while (progress < max_progress) {
     condition_variable.wait(lock);
     seconds elapsed_time = duration_cast<seconds>(steady_clock::now() - last_time);
